/* /////////////////////////////////////////////////////////////////////////
 * File:        pantheios_core.cpp
 *
 * Purpose:     Implementation file for Pantheios core API.
 *
 * Created:     21st June 2005
 * Updated:     15th January 2007
 *
 * Author:      Matthew Wilson
 *
 * Home:        http://www.pantheios.org/
 *
 * Copyright:   Matthew Wilson and Synesis Software Pty Ltd, 1999-2005.
 *              Matthew Wilson, 2005-2006.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the names of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////// */


#include <pantheios/pantheios.h>
#include <pantheios/backend.h>
#include <pantheios/frontend.h>
#include <pantheios/assert.h>

#if defined(STLSOFT_COMPILER_IS_MSVC) && \
            _MSC_VER < 1300
# pragma warning(disable : 4530)
#endif /* _MSC_VER >= 1400 */

#if defined(STLSOFT_COMPILER_IS_MSVC) && \
            _MSC_VER >= 1400
# pragma warning(disable : 4996)
#endif /* _MSC_VER >= 1400 */

#include <stlsoft/stlsoft.h>
#include <stlsoft/iterators/cstring_concatenator_iterator.hpp>
#include <stlsoft/iterators/member_selector_iterator.hpp>
#include <stlsoft/memory/allocator_selector.hpp>
#include <stlsoft/memory/auto_buffer.hpp>
#include <stlsoft/shims/access/string.hpp>
#include <stlsoft/smartptr/scoped_handle.hpp>
#include <stlsoft/synch/lock_scope.hpp>

#include <platformstl/platformstl.h>
//#include <platformstl/synch/atomic_functions.h>
#include <platformstl/synch/spin_mutex.hpp>

#include <new>
#include <numeric>

#include <stdarg.h>
#include <stdio.h>

#ifndef PLATFORMSTL_HAS_ATOMIC_PREINCREMENT
// TODO: Account for ST/MT in non-Windows compilers
# if defined(PLATFORMSTL_OS_IS_UNIX)
#  include <pthread.h>
#  include <platformstl/synch/thread_mutex.hpp>
# else /* ? OS */
#  error Pantheios does not currently support non-UNIX operating systems that do not have support for atomic_preincrement operations. If you have information that will facilitate support for your operation, please contact the Pantheios project members.
# endif /* OS */
#endif /* !PLATFORMSTL_HAS_ATOMIC_PREINCREMENT */

#if defined(PLATFORMSTL_OS_IS_UNIX)
# include <syslog.h>
#endif /* OS */

/* /////////////////////////////////////////////////////////////////////////
 * Compiler compatibility
 */

#if defined(STLSOFT_COMPILER_IS_DMC) || \
    (   defined(STLSOFT_COMPILER_IS_COMO) && \
        defined(_MSC_VER) && \
        defined(WIN32)) || \
    (   defined(STLSOFT_COMPILER_IS_INTEL) && \
        defined(WIN32)) || \
    defined(STLSOFT_COMPILER_IS_MSVC)
# define snprintf           _snprintf
# define vsnprintf          _vsnprintf
#endif /* compiler */

#if (   defined(STLSOFT_COMPILER_IS_MSVC) && \
        _MSC_VER < 1200)
# define _PANTHEIOS_COMPILER_REQUIRES_EXTERNCPP_DEFINITIONS
#endif /* compiler */

/* /////////////////////////////////////////////////////////////////////////
 * Warning suppression
 */

#if defined(STLSOFT_COMPILER_IS_BORLAND)
# pragma warn -8080
#endif /* compiler */

/* /////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(PANTHEIOS_NO_NAMESPACE)
namespace pantheios
{
#endif /* !PANTHEIOS_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////
 * Forward declarations
 */

// Internal function, invoked by core and by external invocations of API
// function pantheios_onBailOut()
static void pantheios_onBailOut_(char const *message, char const *processId);

/* /////////////////////////////////////////////////////////////////////////
 * Constants
 */

const size_t    STACK_BUFFER_SIZE           =   2048;


#if defined(PLATFORMSTL_OS_IS_WIN32)

// The following two are 'known' from pantheios.COM, v1.0.1-b3+. (They will
// *not* be changed).

const DWORD     BAILOUT_1PARAM_MESSAGE_ID   =   1000;
const DWORD     BAILOUT_2PARAM_MESSAGE_ID   =   1001;

#endif /* OS */

/* /////////////////////////////////////////////////////////////////////////
 * Statics
 */

namespace
{
    // API initialisation control variables
    platformstl::sint32_t                       s_apiInit;
    platformstl::spin_mutex::atomic_int_type    s_mutex;
    int                                         s_init;
    int                                         s_isInitialising;

    // API context
    void                                        *s_feToken;
    void                                        *s_beToken;

#if !defined(PLATFORMSTL_HAS_ATOMIC_PREINCREMENT) /* && \
    defined(_REENTRANT) */
    // TODO: This should not be a static, but rather (un)initialised along
    //   with the rest of the (ref-counted) API.
//  pthread_mutex_t                             s_mxBackEndId   =   PTHREAD_MUTEX_INITIALIZER;
    platformstl::thread_mutex                   s_mxBackEndId(false);
#endif /* !PLATFORMSTL_HAS_ATOMIC_PREINCREMENT */

} // anonymous namespace

/* /////////////////////////////////////////////////////////////////////////
 * Core API
 */

namespace
{
    int pantheios_init_onetime()
    {
        PANTHEIOS_ASSERT(!s_isInitialising);

        s_isInitialising = 1;

        int r;

#ifndef PLATFORMSTL_HAS_ATOMIC_PREINCREMENT
        if(NULL == s_mxBackEndId.get())
        {
            r = -3;
        }
        else
#endif /* !PLATFORMSTL_HAS_ATOMIC_PREINCREMENT */
        {
            r   =   pantheios_fe_init(0, &s_feToken);

            if(r < 0)
            {
                pantheios_onBailOut_("front-end did not initialise", pantheios_fe_getProcessIdentity(s_feToken));
            }
            else
            {
                r = pantheios_be_init(pantheios_fe_getProcessIdentity(s_feToken), 0, &s_beToken);

                if(r < 0)
                {
                    pantheios_onBailOut_("back-end did not initialise", pantheios_fe_getProcessIdentity(s_feToken));

                    pantheios_fe_uninit(s_feToken);

                    s_feToken = NULL;
                }
            }
        }

        s_isInitialising = 0;

        return r;
    }

    void pantheios_uninit_onetime()
    {
        pantheios_be_uninit(s_beToken);
        s_beToken = NULL;
        pantheios_fe_uninit(s_feToken);
        s_feToken = NULL;
    }

} // anonymous namespace

/* /////////////////////////////////////////////////////////////////////////
 * Core API
 *
 * Note: for those compilers that object to instantiating templates within 
 * extern "C" functions, the actual functions and their implementations are
 * separated by the use of intermediary extern "C++" forms. For example,
 * pantheios_init() is implemented in terms of pantheios_init__cpp(). For
 * compilers that do not have such a problem, the intermediary forms are not
 * used
 */

#ifdef _PANTHEIOS_COMPILER_REQUIRES_EXTERNCPP_DEFINITIONS
extern "C++" int pantheios_init__cpp();
extern "C++" void pantheios_uninit__cpp();
extern "C++" int pantheios_isSeverityLogged__cpp(pan_sev_t  severity);
extern "C++" int pantheios_log_n__cpp(  pan_sev_t           severity
                                    ,   size_t              numSlices
                                    ,   pan_slice_t const   *slices);
#endif /* _PANTHEIOS_COMPILER_REQUIRES_EXTERNCPP_DEFINITIONS */

PANTHEIOS_CALL(int) pantheios_init()
#ifdef _PANTHEIOS_COMPILER_REQUIRES_EXTERNCPP_DEFINITIONS
{
    return pantheios_init__cpp();
}
int pantheios_init__cpp()
#endif /* _PANTHEIOS_COMPILER_REQUIRES_EXTERNCPP_DEFINITIONS */
{
    platformstl::spin_mutex                         mx(&s_mutex);
    stlsoft::lock_scope<platformstl::spin_mutex>    lock(mx);

    PANTHEIOS_ASSERT(!s_isInitialising);

    if(1 == ++s_apiInit)
    {
        return (s_init = pantheios_init_onetime());
    }
    else
    {
        return s_init;
    }
}

PANTHEIOS_CALL(void) pantheios_uninit()
#ifdef _PANTHEIOS_COMPILER_REQUIRES_EXTERNCPP_DEFINITIONS
{
    pantheios_uninit__cpp();
}
void pantheios_uninit__cpp()
#endif /* _PANTHEIOS_COMPILER_REQUIRES_EXTERNCPP_DEFINITIONS */
{
    platformstl::spin_mutex                         mx(&s_mutex);
    stlsoft::lock_scope<platformstl::spin_mutex>    lock(mx);

    if(0 == --s_apiInit)
    {
        pantheios_uninit_onetime();
    }
}


PANTHEIOS_CALL(int) pantheios_isSeverityLogged(pan_sev_t severity)
#ifdef _PANTHEIOS_COMPILER_REQUIRES_EXTERNCPP_DEFINITIONS
{
    return pantheios_isSeverityLogged__cpp(severity);
}
int pantheios_isSeverityLogged__cpp(pan_sev_t severity)
#endif /* _PANTHEIOS_COMPILER_REQUIRES_EXTERNCPP_DEFINITIONS */
{
    return pantheios_fe_isSeverityLogged(s_feToken, static_cast<int>(severity), 0);
}

PANTHEIOS_CALL(int) pantheios_log_n(    pan_sev_t           severity
                                    ,   size_t              numSlices
                                    ,   pan_slice_t const   *slices)
#ifdef _PANTHEIOS_COMPILER_REQUIRES_EXTERNCPP_DEFINITIONS
{
    return pantheios_log_n__cpp(severity, numSlices, slices);
}
int pantheios_log_n__cpp(   pan_sev_t           severity
                        ,   size_t              numSlices
                        ,   pan_slice_t const   *slices)
#endif /* _PANTHEIOS_COMPILER_REQUIRES_EXTERNCPP_DEFINITIONS */
{
    // Precondition check
    PANTHEIOS_ASSERT(NULL != slices || 0 == numSlices);

    typedef stlsoft::auto_buffer<   char
#ifdef STLSOFT_AUTO_BUFFER_NEW_FORM
                                ,   1 + STACK_BUFFER_SIZE
#endif /* STLSOFT_AUTO_BUFFER_NEW_FORM */
                                >           buffer_t;
#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    try
    {
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */

        // Calculate the total size of the log statement, by summation of the slice array
        const size_t    n   =   std::accumulate(stlsoft::member_selector(slices, &pan_slice_t::len)
                                            ,   stlsoft::member_selector(slices + numSlices, &pan_slice_t::len)
                                            ,   size_t(0));
        buffer_t        buffer(1 + n);
        size_t          nWritten    =   0;

        // We do a check here, so as to cater for cases where the allocator does not
        // throw bad_alloc.
        if(0 == buffer.size())
        {
            return -2;
        }
        else
        {
            // COST: 4KB
            std::copy(  slices
                    ,   slices + numSlices
                    ,   stlsoft::cstring_concatenator(&buffer[0], &nWritten));

            PANTHEIOS_MESSAGE_ASSERT("Written length differs from allocated length", nWritten == n);

            buffer[n] = '\0';

            return pantheios_be_logEntry(s_feToken, s_beToken, static_cast<int>(severity), &buffer[0], n);
        }
#ifdef STLSOFT_CF_EXCEPTION_SUPPORT
    }
    catch(std::bad_alloc &)
    {
        return -2;
    }
    catch(std::exception &)
    {
        return -1;
    }
    catch(...)
    {
        return -3;
    }
#endif /* STLSOFT_CF_EXCEPTION_SUPPORT */

#if (   !defined(STLSOFT_COMPILER_IS_BORLAND) || \
        __BORLANDC__ < 0x0560) && \
    !defined(STLSOFT_COMPILER_IS_COMO) && \
    (   !defined(STLSOFT_COMPILER_IS_MSVC) || \
        _MSC_VER < 1300)
    return 0;
#endif /* compiler */
}

/* /////////////////////////////////////////////////////////////////////////
 * printf() functions
 */

PANTHEIOS_CALL(int) pantheios_printf(   pan_sev_t           severity
                                    ,   char const          *format
                                    ,   ...)
{
    va_list     args;
    int         ret;

    va_start(args, format);

    ret = pantheios_vprintf(severity, format, args);

    va_end(args);

    return ret;
}

PANTHEIOS_CALL(int) pantheios_vprintf(  pan_sev_t           severity
                                    ,   char const          *format
                                    ,   va_list             args)
{
    char    sz[STACK_BUFFER_SIZE + 1];
#if defined(WIN32) && \
    defined(PANTHEIOS_MIN_CRT)
    int     cch =   ::wvsprintfA(&sz[0], format, args);
#else /* ? operating system */
    int     cch =   vsnprintf(&sz[0], STLSOFT_NUM_ELEMENTS(sz) - 1, format, args);
#endif /* operating system */

    sz[STLSOFT_NUM_ELEMENTS(sz) - 1] = '\0';

    return pantheios_be_logEntry(s_feToken, s_beToken, static_cast<int>(severity), &sz[0], static_cast<size_t>(cch));
}

#if 0
PANTHEIOS_CALL(int) pantheios_snprintf( pan_sev_t           severity
                                    ,   char                *dest
                                    ,   size_t              cchDest
                                    ,   char const          *format
                                    ,   ...)
{
    va_list     args;
    int         ret;

    va_start(args, format);

    ret = pantheios_snvprintf(severity, dest, cchDest, format, args);

    va_end(args);

    return ret;
}

PANTHEIOS_CALL(int) pantheios_snvprintf(pan_sev_t           severity
                                    ,   char                *dest
                                    ,   size_t              cchDest
                                    ,   char const          *format
                                    ,   va_list             args)
{
    STLSOFT_SUPPRESS_UNUSED(severity);
    STLSOFT_SUPPRESS_UNUSED(dest);
    STLSOFT_SUPPRESS_UNUSED(cchDest);
    STLSOFT_SUPPRESS_UNUSED(format);
    STLSOFT_SUPPRESS_UNUSED(args);

    return -1;
}
#endif /* 0 */

/* /////////////////////////////////////////////////////////////////////////
 * Utility functions
 */

PANTHEIOS_CALL(void) pantheios_onBailOut(char const *message)
{
    pantheios_onBailOut_(message, (NULL == s_feToken) ? NULL : pantheios_fe_getProcessIdentity(s_feToken));
}

static void pantheios_onBailOut_(char const *message, char const *processId)
{
    if(NULL == message)
    {
        message = "Unspecified failure";
    }

    const size_t    cchMessage  =   ::strlen(message);

#if defined(PLATFORMSTL_OS_IS_WIN32)

    char            message_[2048];
    size_t          cchTotal;
    SYSTEMTIME      st;
    size_t          cchTime;

    ::GetLocalTime(&st);

# ifdef PANTHEIOS_MIN_CRT
    cchTime = static_cast<size_t>(::wsprintf(&message_[0], "%04u%02u%02u-%02u%02u%02u.%03u: ", st.wYear, st.wMonth, st.wDay, st.wHour, st.wMinute, st.wSecond, st.wMilliseconds));
# else /* ? PANTHEIOS_MIN_CRT */
    cchTime = static_cast<size_t>(::sprintf(&message_[0], "%04u%02u%02u-%02u%02u%02u.%03u: ", st.wYear, st.wMonth, st.wDay, st.wHour, st.wMinute, st.wSecond, st.wMilliseconds));
# endif /* PANTHEIOS_MIN_CRT */

    STLSOFT_ASSERT(cchTime < STLSOFT_NUM_ELEMENTS(message_) - 3);

    if((cchTime + cchMessage) < (STLSOFT_NUM_ELEMENTS(message_) - 3))
    {
        cchTotal = cchTime + cchMessage;
    }
    else
    {
        cchTotal = STLSOFT_NUM_ELEMENTS(message_) - 3;
    }

    ::strncpy(&message_[cchTime], message, cchTotal - cchTime);
    message_[cchTotal    ] = '\r';
    message_[cchTotal + 1] = '\n';
    message_[cchTotal + 2] = '\0';

    ////////////////////////////////////
    // 1. Debugger

    ::OutputDebugStringA(message_);

    ////////////////////////////////////
    // 2. Console

    DWORD   numWritten;

    ::WriteFile(::GetStdHandle(STD_ERROR_HANDLE), &message_[0], cchTotal + 2, &numWritten, NULL);

    ////////////////////////////////////
    // 3. File

    HANDLE  hFile = ::CreateFileA(  "logging-bailout.txt"
                                ,   GENERIC_WRITE
                                ,   0
                                ,   NULL
                                ,   OPEN_ALWAYS
                                ,   0
                                ,   NULL);

    if(INVALID_HANDLE_VALUE != hFile)
    {
# if !defined(STLSOFT_COMPILER_IS_MSVC) || \
     _MSC_VER > 1100
        stlsoft::scoped_handle<HANDLE>  hFile_(hFile, ::CloseHandle, INVALID_HANDLE_VALUE);
# endif /* compiler */

        ::SetFilePointer(hFile, 0, NULL, FILE_END);

        ::WriteFile(hFile, &message_[0], cchTotal + 2, &numWritten, NULL);
    }

    ////////////////////////////////////
    // 4. Windows Event Log
    //
    // This uses a special event source registered by
    // Pantheios.COM. If that's not been installed, it
    // still works, but just not as neatly when viewing
    // the event log.
# ifndef PANTHEIOS_MIN_CRT

    HANDLE  hEventSrc = ::RegisterEventSourceA(NULL, "logging-bailout");

    if(NULL != hEventSrc)
    {
#  if !defined(STLSOFT_COMPILER_IS_MSVC) || \
      _MSC_VER > 1100
        stlsoft::scoped_handle<HANDLE>  hEventSrc_(hEventSrc, ::DeregisterEventSource);
#  endif /* compiler */

        LPCTSTR strings[] = 
        {
                processId
            ,   message
        };

        ::ReportEventA( hEventSrc
                    ,   EVENTLOG_ERROR_TYPE
                    ,   0
                    ,   (NULL != processId) ? BAILOUT_2PARAM_MESSAGE_ID : BAILOUT_1PARAM_MESSAGE_ID
                    ,   NULL
                    ,   STLSOFT_NUM_ELEMENTS(strings)
                    ,   0
                    ,   &strings[0]
                    ,   NULL);
        ::GetLastError();
    }
# endif /* PANTHEIOS_MIN_CRT */

#else /* ? OS */

    ////////////////////////////////////
    // 1. Console

    ::fprintf(stderr, "%.*s\n", int(cchMessage), message);

    ////////////////////////////////////
    // 2. File

    FILE    *hFile  =   ::fopen("logging-bailout.txt", "a+");

    if(NULL != hFile)
    {
        stlsoft::scoped_handle<FILE*>   hFile_(hFile, ::fclose);

        ::fprintf(stderr, "%.*s\n", int(cchMessage), message);
    }

# if defined(PLATFORMSTL_OS_IS_UNIX)
    ////////////////////////////////////
    // 3. SysLog

    ::syslog(LOG_EMERG | LOG_USER, "%s", message);
# endif /* OS */

#endif /* OS */
}

PANTHEIOS_CALL(int) pantheios_isInitialising(void)
{
    return s_isInitialising;
}

PANTHEIOS_CALL(int) pantheios_isInitialised(void)
{
    return 0 != s_apiInit;
}

PANTHEIOS_CALL(void) pantheios_exit_process(int code)
{
#if defined(WIN32)

    // By rights, Windows programs should use exit() to close down. However,
    // pantheios_exit_process() is not a general substitute for exit(). It is
    // specifically intended to Get out of Dodge asap, in the rare case where
    // Pantheios cannot be initialised.
    //
    // NOTE: May need to revisit this in the future, in cases where a
    // sophisticated back-end, e.g. log4cxx, does not initialise.

    ::ExitProcess(static_cast<DWORD>(code));

#else /* ? operating system */

     ::exit(code);

#endif /* operating system */
}

PANTHEIOS_CALL(int) pantheios_getNextBackEndId(void)
{
#ifdef PLATFORMSTL_HAS_ATOMIC_PREINCREMENT

    static platformstl::atomic_int_t    s_nextId    =   1000;

    return static_cast<int>(platformstl::atomic_preincrement(&s_nextId));

#else /* ? PLATFORMSTL_HAS_ATOMIC_PREINCREMENT */

    static int							s_nextId    =   1000;

//  platformstl::thread_mutex                       mx(&s_mxBackEndId, false);  // "Borrow" the mutex
//  stlsoft::lock_scope<platformstl::thread_mutex>  lock(mx);

    stlsoft::lock_scope<platformstl::thread_mutex>  lock(s_mxBackEndId);

    return ++s_nextId;
#endif /* PLATFORMSTL_HAS_ATOMIC_PREINCREMENT */
}

/* static */ size_t pan_slice_t::get_lazy_length(size_t fromLen, size_t toLen)
{
    const size_t topBit     =   size_t(0x01) << (sizeof(size_t) * 8 - 1);

    return topBit | (fromLen & 0x7fff) | ((toLen & 0x7fff) << 15);
}

PANTHEIOS_CALL(size_t) pantheios_strnlen(char const *s, size_t len)
{
    const size_t topBit     =   size_t(0x01) << (sizeof(size_t) * 8 - 1);
    const size_t nextTopBit =   size_t(0x01) << (sizeof(size_t) * 8 - 2);

    if(topBit & len)
    {
        // Top bit is set. This can mean one of two things:
        //
        // 1. strlen() is requested. This is indicated by next top bit
        //    being set (e.g. if len == size_t(-1)
        // 2. min/max strlen is requested. This is indicated by next top
        //    bit being not set.

        if(nextTopBit & len)
        {
            // 1. strlen() is requested.
            return (NULL != s) ? ::strlen(s) : 0;
        }
        else
        {
            // 2. min/max strlen is requested.

            // This requires some smarts

            // fromLen is lowest 15 bits
            const size_t    fromLen =   len & 0x7fff;

            // toLen is bit16-bit30
            const size_t    toLen   =   (len >> 15)  & 0x7fff;

            char const      *from   =   s + fromLen;
            char const      *to     =   s + toLen;
            char const      *nul;

            if(fromLen < toLen)
            {
                // Search forwards
                nul = std::find(from
                            ,   to + 1  // +1 because search is inclusive
                            ,   '\0');
            }
            else
            {
                nul = from;

                // Search backwards
                do
                {
                    if('\0' == *from)
                    {
                        nul = from;

                        break;
                    }
                }
                while(--from != to);
            }

            return static_cast<size_t>(nul - s);
        }
    }
    else
    {
        // No top bit set, so nothing special. Just return
        // the length as given.

        return len;
    }
}

PANTHEIOS_CALL(char *)  pantheios_strdup_nothrow(char const *s)
{
    if(NULL == s)
    {
        return NULL;
    }
    else
    {
        const size_t    len =   ::strlen(s);
        char            *s2 =   static_cast<char*>(::malloc(sizeof(char) * (1 + len)));

        if(NULL != s2)
        {
            ::memcpy(s2, s, sizeof(char) * (1 + len));
        }

        return s2;
    }
}

PANTHEIOS_CPP_CALL(char *)  pantheios_strdup_throw(char const *s)
{
    char    *s2 = pantheios_strdup_nothrow(s);

    if( NULL == s2 &&
        NULL != s)
    {
#ifdef STLSOFT_CF_THROW_BAD_ALLOC

        throw std::bad_alloc();

#endif /* STLSOFT_CF_THROW_BAD_ALLOC */
    }

    return s2;
}

PANTHEIOS_CALL(void) pantheios_strfree(char *s)
{
    ::free(s);
}

PANTHEIOS_CALL(void) pantheios_puts(pan_sev_t severity, char const *s)
{
    pan_slice_t slice(-1, s);

    pantheios_log_n(severity, 1, &slice);
}


#if defined(STLSOFT_COMPILER_IS_MSVC) && \
            _MSC_VER >= 1400
#else /* ? "safe functions" */
# if defined(STLSOFT_COMPILER_IS_DMC) || \
     defined(STLSOFT_COMPILER_IS_MSVC) || \
     (   defined(STLSOFT_COMPILER_IS_INTEL) && \
         defined(WIN32))
# if defined(PANTHEIOS_MIN_CRT)
extern "C" size_t panthieos_inserter_snprintf_(char *dest, size_t cchDest, const char *format, ...)
{
    va_list     args;
    int         ret;

    PANTHEIOS_ASSERT(cchDest < 256);

    va_start(args, format);

    ret = ::wvsprintfA(dest, format, args);

    va_end(args);

    return ret;
}
# else /* ? PANTHEIOS_MIN_CRT */
# endif /* PANTHEIOS_MIN_CRT */
# else /* ? compiler */
# endif /* compiler */
#endif /* "safe functions" */

/* /////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(PANTHEIOS_NO_NAMESPACE)
} /* namespace pantheios */
#endif /* !PANTHEIOS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////// */
